@@ -222,7 +222,7 @@ on_server_close_finish (GObject        *object,
 
   g_io_stream_close_finish (G_IO_STREAM (object), res, &error);
   if (test->expect_server_error)
-    g_assert (error != NULL);
+    g_assert_nonnull (error);
   else
     g_assert_no_error (error);
   test->server_running = FALSE;
@@ -242,7 +242,7 @@ on_output_write_finish (GObject        *object,
 {
   TestConnection *test = user_data;
 
-  g_assert (test->server_error == NULL);
+  g_assert_no_error (test->server_error);
   g_output_stream_write_finish (G_OUTPUT_STREAM (object), res, &test->server_error);
 
   if (!test->server_error && test->rehandshake)
@@ -451,7 +451,7 @@ on_input_read_finish (GObject        *object,
                                                NULL, &test->read_error);
   if (!test->read_error)
     {
-      g_assert (line);
+      g_assert_nonnull (line);
 
       check = g_strdup (TEST_DATA);
       g_strstrip (check);
@@ -470,7 +470,7 @@ read_test_data_async (TestConnection *test)
   GDataInputStream *stream;
 
   stream = g_data_input_stream_new (g_io_stream_get_input_stream (test->client_connection));
-  g_assert (stream);
+  g_assert_nonnull (stream);
 
   g_data_input_stream_read_line_async (stream, G_PRIORITY_DEFAULT, NULL,
                                        on_input_read_finish, test);
@@ -509,12 +509,12 @@ test_verified_connection (TestConnection *test,
 
   test->database = g_tls_file_database_new (tls_test_file_path ("ca-roots.pem"), &error);
   g_assert_no_error (error);
-  g_assert (test->database);
+  g_assert_nonnull (test->database);
 
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_NONE, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
@@ -546,19 +546,19 @@ test_verified_chain (TestConnection *test,
   /* Prepare the intermediate cert. */
   intermediate_cert = g_tls_certificate_new_from_file (tls_test_file_path ("intermediate-ca.pem"), &error);
   g_assert_no_error (error);
-  g_assert (intermediate_cert);
+  g_assert_nonnull (intermediate_cert);
 
   /* Prepare the server cert. */
   g_clear_pointer (&cert_data, g_free);
   g_file_get_contents (tls_test_file_path ("server-intermediate.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   g_file_get_contents (tls_test_file_path ("server-intermediate-key.pem"),
                        &key_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (key_data);
+  g_assert_nonnull (key_data);
 
   server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                 NULL, &error,
@@ -567,7 +567,7 @@ test_verified_chain (TestConnection *test,
                                 "private-key-pem", key_data,
                                 NULL);
   g_assert_no_error (error);
-  g_assert (server_cert);
+  g_assert_nonnull (server_cert);
 
   g_object_unref (intermediate_cert);
   g_free (cert_data);
@@ -594,13 +594,13 @@ test_verified_chain_with_redundant_root_cert (TestConnection *test,
   /* The root is redundant. It should not hurt anything. */
   root_cert = g_tls_certificate_new_from_file (tls_test_file_path ("ca.pem"), &error);
   g_assert_no_error (error);
-  g_assert (root_cert);
+  g_assert_nonnull (root_cert);
 
   /* Prepare the intermediate cert. */
   g_file_get_contents (tls_test_file_path ("intermediate-ca.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   intermediate_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                       NULL, &error,
@@ -608,19 +608,19 @@ test_verified_chain_with_redundant_root_cert (TestConnection *test,
                                       "certificate-pem", cert_data,
                                       NULL);
   g_assert_no_error (error);
-  g_assert (intermediate_cert);
+  g_assert_nonnull (intermediate_cert);
 
   /* Prepare the server cert. */
   g_clear_pointer (&cert_data, g_free);
   g_file_get_contents (tls_test_file_path ("server-intermediate.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   g_file_get_contents (tls_test_file_path ("server-intermediate-key.pem"),
                        &key_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (key_data);
+  g_assert_nonnull (key_data);
 
   server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                 NULL, &error,
@@ -629,7 +629,7 @@ test_verified_chain_with_redundant_root_cert (TestConnection *test,
                                 "private-key-pem", key_data,
                                 NULL);
   g_assert_no_error (error);
-  g_assert (server_cert);
+  g_assert_nonnull (server_cert);
 
   g_object_unref (intermediate_cert);
   g_object_unref (root_cert);
@@ -662,19 +662,19 @@ test_verified_chain_with_duplicate_server_cert (TestConnection *test,
   /* Prepare the intermediate cert. */
   intermediate_cert = g_tls_certificate_new_from_file (tls_test_file_path ("intermediate-ca.pem"), &error);
   g_assert_no_error (error);
-  g_assert (intermediate_cert);
+  g_assert_nonnull (intermediate_cert);
 
   /* Prepare the server cert. */
   g_clear_pointer (&cert_data, g_free);
   g_file_get_contents (tls_test_file_path ("server-intermediate.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   g_file_get_contents (tls_test_file_path ("server-intermediate-key.pem"),
                        &key_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (key_data);
+  g_assert_nonnull (key_data);
 
   server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                 NULL, &error,
@@ -682,7 +682,7 @@ test_verified_chain_with_duplicate_server_cert (TestConnection *test,
                                 "certificate-pem", cert_data,
                                 NULL);
   g_assert_no_error (error);
-  g_assert (server_cert);
+  g_assert_nonnull (server_cert);
 
   /* Prepare the server cert... again. Private key must go on this one. */
   extra_server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
@@ -692,7 +692,7 @@ test_verified_chain_with_duplicate_server_cert (TestConnection *test,
                                       "private-key-pem", key_data,
                                       NULL);
   g_assert_no_error (error);
-  g_assert (extra_server_cert);
+  g_assert_nonnull (extra_server_cert);
 
   g_object_unref (intermediate_cert);
   g_object_unref (server_cert);
@@ -721,11 +721,11 @@ test_verified_unordered_chain (TestConnection *test,
   intermediate_cert = g_tls_certificate_new_from_file (tls_test_file_path ("intermediate-ca.pem"),
                                                        &error);
   g_assert_no_error (error);
-  g_assert (intermediate_cert);
+  g_assert_nonnull (intermediate_cert);
 
   g_file_get_contents (tls_test_file_path ("ca.pem"), &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   /* Prepare the root cert (to be sent in the middle of the chain). */
   root_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
@@ -734,18 +734,18 @@ test_verified_unordered_chain (TestConnection *test,
                               "certificate-pem", cert_data,
                               NULL);
   g_assert_no_error (error);
-  g_assert (root_cert);
+  g_assert_nonnull (root_cert);
 
   g_clear_pointer (&cert_data, g_free);
   g_file_get_contents (tls_test_file_path ("server-intermediate.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   g_file_get_contents (tls_test_file_path ("server-intermediate-key.pem"),
                        &key_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (key_data);
+  g_assert_nonnull (key_data);
 
   /* Prepare the server cert. */
   server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
@@ -755,7 +755,7 @@ test_verified_unordered_chain (TestConnection *test,
                                 "private-key-pem", key_data,
                                 NULL);
   g_assert_no_error (error);
-  g_assert (server_cert);
+  g_assert_nonnull (server_cert);
 
   g_object_unref (intermediate_cert);
   g_object_unref (root_cert);
@@ -786,7 +786,7 @@ test_verified_chain_with_alternative_ca_cert (TestConnection *test,
    * fail, since the issuer is untrusted. */
   root_cert = g_tls_certificate_new_from_file (tls_test_file_path ("ca-alternative.pem"), &error);
   g_assert_no_error (error);
-  g_assert (root_cert);
+  g_assert_nonnull (root_cert);
 
   /* Prepare the intermediate cert. Modern TLS libraries are expected to notice
    * that it is signed by the same public key as a certificate in the root
@@ -797,7 +797,7 @@ test_verified_chain_with_alternative_ca_cert (TestConnection *test,
   g_file_get_contents (tls_test_file_path ("intermediate-ca.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   intermediate_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                       NULL, &error,
@@ -805,19 +805,19 @@ test_verified_chain_with_alternative_ca_cert (TestConnection *test,
                                       "certificate-pem", cert_data,
                                       NULL);
   g_assert_no_error (error);
-  g_assert (intermediate_cert);
+  g_assert_nonnull (intermediate_cert);
 
   /* Prepare the server cert. */
   g_clear_pointer (&cert_data, g_free);
   g_file_get_contents (tls_test_file_path ("server-intermediate.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   g_file_get_contents (tls_test_file_path ("server-intermediate-key.pem"),
                        &key_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (key_data);
+  g_assert_nonnull (key_data);
 
   server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                 NULL, &error,
@@ -826,7 +826,7 @@ test_verified_chain_with_alternative_ca_cert (TestConnection *test,
                                 "private-key-pem", key_data,
                                 NULL);
   g_assert_no_error (error);
-  g_assert (server_cert);
+  g_assert_nonnull (server_cert);
 
   g_object_unref (intermediate_cert);
   g_object_unref (root_cert);
@@ -854,19 +854,19 @@ test_invalid_chain_with_alternative_ca_cert (TestConnection *test,
   /* This certificate has the same public key as a certificate in the root store. */
   root_cert = g_tls_certificate_new_from_file (tls_test_file_path ("ca-alternative.pem"), &error);
   g_assert_no_error (error);
-  g_assert (root_cert);
+  g_assert_nonnull (root_cert);
 
   /* The intermediate cert is not sent. The chain should be rejected, since without intermediate.pem
    * there is no proof that ca-alternative.pem signed server-intermediate.pem. */
   g_file_get_contents (tls_test_file_path ("server-intermediate.pem"),
                        &cert_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (cert_data);
+  g_assert_nonnull (cert_data);
 
   g_file_get_contents (tls_test_file_path ("server-intermediate-key.pem"),
                        &key_data, NULL, &error);
   g_assert_no_error (error);
-  g_assert (key_data);
+  g_assert_nonnull (key_data);
 
   server_cert = g_initable_new (g_tls_backend_get_certificate_type (backend),
                                 NULL, &error,
@@ -875,7 +875,7 @@ test_invalid_chain_with_alternative_ca_cert (TestConnection *test,
                                 "private-key-pem", key_data,
                                 NULL);
   g_assert_no_error (error);
-  g_assert (server_cert);
+  g_assert_nonnull (server_cert);
 
   g_object_unref (root_cert);
   g_free (cert_data);
@@ -885,7 +885,7 @@ test_invalid_chain_with_alternative_ca_cert (TestConnection *test,
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_NONE, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
@@ -907,7 +907,7 @@ on_notify_accepted_cas (GObject *obj,
                         gpointer user_data)
 {
   gboolean *changed = user_data;
-  g_assert (*changed == FALSE);
+  g_assert_false (*changed);
   *changed = TRUE;
 }
 
@@ -924,12 +924,12 @@ test_client_auth_connection (TestConnection *test,
 
   test->database = g_tls_file_database_new (tls_test_file_path ("ca-roots.pem"), &error);
   g_assert_no_error (error);
-  g_assert (test->database);
+  g_assert_nonnull (test->database);
 
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_REQUIRED, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
@@ -954,9 +954,9 @@ test_client_auth_connection (TestConnection *test,
   g_assert_no_error (test->server_error);
 
   peer = g_tls_connection_get_peer_certificate (G_TLS_CONNECTION (test->server_connection));
-  g_assert (peer != NULL);
-  g_assert (g_tls_certificate_is_same (peer, cert));
-  g_assert (cas_changed == TRUE);
+  g_assert_nonnull (peer);
+  g_assert_true (g_tls_certificate_is_same (peer, cert));
+  g_assert_true (cas_changed);
 
   g_object_unref (cert);
   g_object_unref (test->database);
@@ -987,8 +987,8 @@ test_client_auth_connection (TestConnection *test,
 
   /* peer should see the second client cert */
   peer = g_tls_connection_get_peer_certificate (G_TLS_CONNECTION (test->server_connection));
-  g_assert (peer != NULL);
-  g_assert (g_tls_certificate_is_same (peer, cert));
+  g_assert_nonnull (peer);
+  g_assert_true (g_tls_certificate_is_same (peer, cert));
 }
 
 static void
@@ -1013,12 +1013,12 @@ test_client_auth_failure (TestConnection *test,
 
   test->database = g_tls_file_database_new (tls_test_file_path ("ca-roots.pem"), &error);
   g_assert_no_error (error);
-  g_assert (test->database);
+  g_assert_nonnull (test->database);
 
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_REQUIRED, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
@@ -1039,7 +1039,7 @@ test_client_auth_failure (TestConnection *test,
   g_assert_error (test->read_error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
   g_assert_error (test->server_error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
 
-  g_assert (accepted_changed == TRUE);
+  g_assert_true (accepted_changed);
 
   g_object_unref (test->client_connection);
   g_object_unref (test->database);
@@ -1080,13 +1080,54 @@ test_client_auth_failure (TestConnection *test,
   g_assert_no_error (test->server_error);
 
   peer = g_tls_connection_get_peer_certificate (G_TLS_CONNECTION (test->server_connection));
-  g_assert (peer != NULL);
-  g_assert (g_tls_certificate_is_same (peer, cert));
-  g_assert (accepted_changed == TRUE);
+  g_assert_nonnull (peer);
+  g_assert_true (g_tls_certificate_is_same (peer, cert));
+  g_assert_true (accepted_changed);
 
   g_object_unref (cert);
 }
 
+static void
+test_client_auth_fail_missing_client_private_key (TestConnection *test,
+                                                  gconstpointer   data)
+{
+  GTlsCertificate *cert;
+  GIOStream *connection;
+  GError *error = NULL;
+
+  g_test_bug ("793712");
+
+  test->database = g_tls_file_database_new (tls_test_file_path ("ca-roots.pem"), &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (test->database);
+
+  connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_REQUIRED, TRUE);
+  test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (test->client_connection);
+  g_object_unref (connection);
+
+  g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
+
+  /* Oops: we "accidentally" set client.pem rather than client-and-key.pem. The
+   * connection will fail, but we should not crash.
+   */
+  cert = g_tls_certificate_new_from_file (tls_test_file_path ("client.pem"), &error);
+  g_assert_no_error (error);
+
+  g_tls_connection_set_certificate (G_TLS_CONNECTION (test->client_connection), cert);
+
+  /* All validation in this test */
+  g_tls_client_connection_set_validation_flags (G_TLS_CLIENT_CONNECTION (test->client_connection),
+                                                G_TLS_CERTIFICATE_VALIDATE_ALL);
+
+  read_test_data_async (test);
+  g_main_loop_run (test->loop);
+
+  g_assert_error (test->read_error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
+  g_assert_no_error (test->server_error);
+}
+
 static void
 test_client_auth_request_cert (TestConnection *test,
                                gconstpointer   data)
@@ -1100,12 +1141,12 @@ test_client_auth_request_cert (TestConnection *test,
 
   test->database = g_tls_file_database_new (tls_test_file_path ("ca-roots.pem"), &error);
   g_assert_no_error (error);
-  g_assert (test->database);
+  g_assert_nonnull (test->database);
 
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_REQUIRED, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
@@ -1132,9 +1173,9 @@ test_client_auth_request_cert (TestConnection *test,
   g_assert_no_error (test->server_error);
 
   peer = g_tls_connection_get_peer_certificate (G_TLS_CONNECTION (test->server_connection));
-  g_assert (peer != NULL);
-  g_assert (g_tls_certificate_is_same (peer, cert));
-  g_assert (cas_changed == TRUE);
+  g_assert_nonnull (peer);
+  g_assert_true (g_tls_certificate_is_same (peer, cert));
+  g_assert_true (cas_changed);
 
   g_object_unref (cert);
 }
@@ -1149,12 +1190,12 @@ test_client_auth_request_fail (TestConnection *test,
 
   test->database = g_tls_file_database_new (tls_test_file_path ("ca-roots.pem"), &error);
   g_assert_no_error (error);
-  g_assert (test->database);
+  g_assert_nonnull (test->database);
 
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_REQUIRED, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
@@ -1187,7 +1228,7 @@ test_connection_no_database (TestConnection *test,
   connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_NONE, TRUE);
   test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
   g_assert_no_error (error);
-  g_assert (test->client_connection);
+  g_assert_nonnull (test->client_connection);
   g_object_unref (connection);
 
   /* Overrides loading of the default database */
@@ -1297,9 +1338,9 @@ test_connection_socket_client (TestConnection *test,
   connection = (GSocketConnection *)test->client_connection;
   test->client_connection = NULL;
 
-  g_assert (G_IS_TCP_WRAPPER_CONNECTION (connection));
+  g_assert_true (G_IS_TCP_WRAPPER_CONNECTION (connection));
   base = g_tcp_wrapper_connection_get_base_io_stream (G_TCP_WRAPPER_CONNECTION (connection));
-  g_assert (G_IS_TLS_CONNECTION (base));
+  g_assert_true (G_IS_TLS_CONNECTION (base));
 
   g_io_stream_close (G_IO_STREAM (connection), NULL, &error);
   g_assert_no_error (error);
@@ -1417,9 +1458,9 @@ test_connection_read_time_out_write (TestConnection *test,
   connection = (GSocketConnection *)test->client_connection;
   test->client_connection = NULL;
 
-  g_assert (G_IS_TCP_WRAPPER_CONNECTION (connection));
+  g_assert_true (G_IS_TCP_WRAPPER_CONNECTION (connection));
   base = g_tcp_wrapper_connection_get_base_io_stream (G_TCP_WRAPPER_CONNECTION (connection));
-  g_assert (G_IS_TLS_CONNECTION (base));
+  g_assert_true (G_IS_TLS_CONNECTION (base));
 
   g_io_stream_close (G_IO_STREAM (connection), NULL, &error);
   g_assert_no_error (error);
@@ -2009,20 +2050,19 @@ test_output_stream_close (TestConnection *test,
     g_main_context_iteration (NULL, TRUE);
 
   ret = g_output_stream_close (g_io_stream_get_output_stream (test->client_connection),
-      NULL, &error);
+                               NULL, &error);
   g_assert_no_error (error);
-  g_assert (ret);
-
+  g_assert_true (ret);
 
   /* Verify that double close returns TRUE */
   ret = g_output_stream_close (g_io_stream_get_output_stream (test->client_connection),
-      NULL, &error);
+                               NULL, &error);
   g_assert_no_error (error);
-  g_assert (ret);
+  g_assert_true (ret);
 
   size = g_output_stream_write (g_io_stream_get_output_stream (test->client_connection),
                                 "data", 4, NULL, &error);
-  g_assert (size == -1);
+  g_assert_cmpint (size, ==, -1);
   g_assert_error (error, G_IO_ERROR, G_IO_ERROR_CLOSED);
   g_clear_error (&error);
 
@@ -2037,7 +2077,84 @@ test_output_stream_close (TestConnection *test,
 
   ret = g_io_stream_close (test->client_connection, NULL, &error);
   g_assert_no_error (error);
-  g_assert (ret);
+  g_assert_true (ret);
+}
+
+static void
+test_garbage_database (TestConnection *test,
+                       gconstpointer   data)
+{
+  GIOStream *connection;
+  GError *error = NULL;
+
+  test->database = g_tls_file_database_new (tls_test_file_path ("garbage.pem"), &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (test->database);
+
+  connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_NONE, TRUE);
+  test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
+  g_assert_no_error (error);
+  g_assert_nonnull (test->client_connection);
+  g_object_unref (connection);
+
+  g_tls_connection_set_database (G_TLS_CONNECTION (test->client_connection), test->database);
+
+  /* All validation in this test */
+  g_tls_client_connection_set_validation_flags (G_TLS_CLIENT_CONNECTION (test->client_connection),
+                                                G_TLS_CERTIFICATE_VALIDATE_ALL);
+
+  read_test_data_async (test);
+  g_main_loop_run (test->loop);
+
+  /* Should reject the server's certificate, because our TLS database contains
+   * no valid certificates.
+   */
+  g_assert_error (test->read_error, G_TLS_ERROR, G_TLS_ERROR_BAD_CERTIFICATE);
+  g_assert_no_error (test->server_error);
+}
+
+static void
+test_readwrite_after_connection_destroyed (TestConnection *test,
+                                           gconstpointer   data)
+{
+  GIOStream *connection;
+  GOutputStream *ostream;
+  GInputStream *istream;
+  unsigned char buffer[1];
+  GError *error = NULL;
+
+  g_test_bug ("792219");
+
+  connection = start_async_server_and_connect_to_it (test, G_TLS_AUTHENTICATION_REQUESTED, TRUE);
+  test->client_connection = g_tls_client_connection_new (connection, test->identity, &error);
+  g_assert_no_error (error);
+  g_object_unref (connection);
+
+  istream = g_object_ref (g_io_stream_get_input_stream (test->client_connection));
+  ostream = g_object_ref (g_io_stream_get_output_stream (test->client_connection));
+  g_clear_object (&test->client_connection);
+
+  /* The GTlsConnection has been destroyed, but its underlying streams
+   * live on, because we have reffed them. Verify that attempts to read
+   * and write produce only nice GErrors.
+   */
+  g_input_stream_read (istream, buffer, sizeof (buffer), NULL, &error);
+  g_assert_error (error, G_IO_ERROR, G_IO_ERROR_CLOSED);
+  g_clear_error (&error);
+
+  g_output_stream_write (ostream, TEST_DATA, TEST_DATA_LENGTH,
+                         G_PRIORITY_DEFAULT, &error);
+  g_assert_error (error, G_IO_ERROR, G_IO_ERROR_CLOSED);
+  g_clear_error (&error);
+
+  g_input_stream_close (istream, NULL, &error);
+  g_assert_no_error (error);
+
+  g_output_stream_close (ostream, NULL, &error);
+  g_assert_no_error (error);
+
+  g_object_unref (istream);
+  g_object_unref (ostream);
 }
 
 int
@@ -2050,7 +2167,6 @@ main (int   argc,
   g_test_bug_base ("http://bugzilla.gnome.org/");
 
   g_setenv ("GSETTINGS_BACKEND", "memory", TRUE);
-  g_setenv ("GIO_EXTRA_MODULES", TOP_BUILDDIR "/tls/gnutls/.libs", TRUE);
   g_setenv ("GIO_USE_TLS", "gnutls", TRUE);
 
   g_test_add ("/tls/connection/basic", TestConnection, NULL,
@@ -2075,6 +2191,8 @@ main (int   argc,
               setup_connection, test_client_auth_rehandshake, teardown_connection);
   g_test_add ("/tls/connection/client-auth-failure", TestConnection, NULL,
               setup_connection, test_client_auth_failure, teardown_connection);
+  g_test_add ("/tls/connection/client-auth-fail-missing-client-private-key", TestConnection, NULL,
+              setup_connection, test_client_auth_fail_missing_client_private_key, teardown_connection);
   g_test_add ("/tls/connection/client-auth-request-cert", TestConnection, NULL,
               setup_connection, test_client_auth_request_cert, teardown_connection);
   g_test_add ("/tls/connection/client-auth-request-fail", TestConnection, NULL,
@@ -2111,8 +2229,12 @@ main (int   argc,
               setup_connection, test_output_stream_close, teardown_connection);
   g_test_add ("/tls/connection/fallback", TestConnection, NULL,
               setup_connection, test_fallback, teardown_connection);
+  g_test_add ("/tls/connection/garbage-database", TestConnection, NULL,
+              setup_connection, test_garbage_database, teardown_connection);
+  g_test_add ("/tls/connection/readwrite-after-connection-destroyed", TestConnection, NULL,
+              setup_connection, test_readwrite_after_connection_destroyed, teardown_connection);
 
-  ret = g_test_run();
+  ret = g_test_run ();
 
   /* for valgrinding */
   g_main_context_unref (g_main_context_default ());
